// Sum array elements using System V threads (fork/wait)
// Try "gcc sum_test.c" -> compile and "./a.out" -> run
// To read more, see
// a) Chapter 15.9 from W. Richard Stevens and Stephan A. Rago, 
// "Advanced Programming in the Unix Environment", Addison Wesley, 2014, 3rd edition.
// b) http://www.yolinux.com/TUTORIALS/ForkExecProcesses.html
//
// Also detailed description in Richard Stevens, Unix Network
// Programming, Volume II (pages 352-356 for System V shared memory)
// (Volume I on sockets -- is also interesting)

// Can you extend the code to work with more threads
// Can you view the statistics from third arg of shmctl function (struct shmid-ds)

#include <sys/ipc.h>    // interprocess communication header
#include <unistd.h>     // fork()
#include <sys/wait.h>   // wait(0)
#include <sys/shm.h>    // shared memory

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>

#define MAX_NUMBER 10	// random numbers in [0 .. MAX_NUMBER-1]
#define N 10000		// array size

/* Function to print shmem stats at removal
   */
void print_shm_stats(int shmid, struct shmid_ds *buf)
{
  	printf("segment ID: %d \n", shmid);
  	printf("uid: %d \n", buf->shm_perm.uid);  
  	printf("gid: %d \n", buf->shm_perm.gid);  
  	printf("cuid: %d \n", buf->shm_perm.cuid);  
  	printf("cgid: %d \n", buf->shm_perm.cgid);  
  	printf("Mode: %d \n", buf->shm_perm.mode);
  	printf("Size: %zu \n", buf->shm_segsz);
  	printf("Pid of last shmop: %d \n", (int)buf->shm_lpid);
  	printf("Pid of creator: %d \n", (int)buf->shm_cpid);
  	printf("No. of attaches: %d \n", (int)buf->shm_nattch);
  	printf("The last shmat time: %ld \n", (int)buf->shm_atime);
  	printf("The last shmdt time: %ld \n", (int)buf->shm_dtime);
  	printf("The last change time: %ld \n", (int)buf->shm_ctime);
}

/* Function that fills array a[] with integer random numbers
   */
int readmat(int a[],int n)
{
	long i;

	if (n > 0) {
		for (i=0;i<n;i++)
			a[i]= rand() % MAX_NUMBER;
		return(0);
	}
	return(-1);
}

/* Function that prints array a[]
   */
int writemat(int a[],int n)
{
	long i;

	if (n > 0) {
		for (i=0;i<n;i++)
			printf(" a[%d] = %d \n", i, a[i]);
		return(0);
	}
	return(-1);
}

int main()
{
	// array with elements
	int a[N];
	int sum=0, i;
	int *ptr;
	int myid, shmid;
	struct shmid_ds buf;

	int pfork(int);
	void pjoin(int ,int);


	srand(1L);
     	if(readmat(a, N) < 0) {
		printf("Array size is negative \n");
		exit(1);
	}

	writemat(a, N);

	// allocate a physical memory segment: pointer to physical memory
	// (can check with ipcs command)
	shmid=shmget(IPC_PRIVATE,4,IPC_CREAT|0666);
	if(shmid < 0)
		printf("ERROR\n");
	else
	printf("Shmid Is %d\n",shmid);
    
	// attach physical memory segment to process virtual address space 
	// (to get virtual address pointer)
	ptr=shmat(shmid,0,0);

	// initialize value of virtual address pointer
	*ptr=0;    

	// create a child process
	myid=pfork(1);

	if(myid==0) { // parent
		// sum every two terms starting from zero array element
		for(i=0;i<N;i+=2)
		{
			sum=sum+a[i];
		}
		printf("Sum in parent is %d\n",sum);
	}
	else if(myid==1) { // child with myid = 1
		// sum every two terms starting from first array element
		for(i=1;i<N;i+=2)
		{
				*ptr=*ptr+a[i];
		}
		printf("Sum in child is %d\n",*ptr);
	}

	// wait for termination before accessing sum and *ptr
	pjoin(1,myid);

	sum=sum+*ptr;

	//system("sleep 20");
	//shmctl(shmid, IPC_RMID, NULL);
	if(shmctl(shmid, IPC_STAT, &buf) == -1){
		perror("shmctl");
		exit(EXIT_FAILURE);
	}

	print_shm_stats(shmid, &buf);

        shmctl(shmid, IPC_RMID, NULL);

	printf("Total sum is %d\n",sum);
	return 0;
}

int pfork(int x)
{
	int j;

	for (j=1;j<=x;j++) {
		if(fork()==0)
			return j;
	}
	return 0;
}

void pjoin(int x,int id)
{
int j;

	if (id==0) {
		for(j=1;j<=x;j++) {
			wait(0);
		}    
	}
	else {
		exit(0);
	}
}
