// Simple program to compute dot product with many pthreads
// See from Stevens and Rago book, Sections 11.1 to 11.6.1
// (pages 383-401) with an emphasis on Sections 11.4,
// 11.5 (pthread_create and pthread_join) and 11.6.1 
// (pthread_mutex_init kai  pthread_mutex_lock, pthread_mutex_unlock).
//
// Also useful in Section 12.4.1 in case threads are related to different processes.
//
// Examine the use of locks, i.e.
// a) definitions of mutex using pthread_mutex_t,
// b) initialization using pthread_mutex_init, and
// c) use of pthread_mutex_lock/unlock for exclusive access
//    to DOTPRODUCT which is written by many (actually all) pthreads
// Remarks: 
// If (c) is omitted the result is wrong (try to examine this)
// 
// Rewrite to perform matrix multiplication, i.e. C = A x B (A, B, C Matrices)
// Can you rewrite to create a data race hazard? How?


#define _GNU_SOURCE     // required by pthread_yield (do: man pthread_yield)

#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/times.h>
#include <unistd.h> 

#define NO_ITERATIONS 100	// use many runs for better timing accuracy
#define MAX_NUMBER   1000	// matrix with random integers (0 to 9)
#define N    100		// matrix size
#define NPROC 100		// number of child processes (there is no parent)

pthread_t dotThread[NPROC]; // define the NPROC threads
void *DotProduct (void *); // threads use this function call

// mutex for protecting DOTPRODUCT which is written by all pthreads
pthread_mutex_t dotsum;

int stripSize; // variable used for data allocation
long i;

// Structure with shared variables (used by all pthreads)
struct whatever{
  int A[N];
  int B[N];
  int DOTPRODUCT;
}*myvars;

// Function that reads numbers
int read_vector(int vector[N],int n) {
	long i;
    
	for(i=0;i<n;i++){

		//vector[i]= (int) (MAX_NUMBER*(float) rand()/(RAND_MAX+1.0));

		// value 1 for easy testing
		vector[i]= 1;
	}

	return(0);
}  

int main(int argc,char *argv[])
{
      
	int myid=0,no_runs;

	// compute average timing
	double ticspersec;
	double t1,t2,avrg_time=0.0;
	double sum_avrg_time_all_iterations=0;
	int no_iterations,delay;
	struct tms tb;

	pthread_attr_t attr;
   
	srand(1L); 

	for (no_runs=0; no_runs<NO_ITERATIONS; no_runs++) {

		// allocation of shared memory (used by all threads)
		myvars = (struct whatever*)malloc (NPROC*sizeof(struct whatever));

		// read array A
		if(read_vector((int *)myvars->A,N)<0) {
			printf("File has negative arguments.\n");
			exit(0);
		}

		// read array B
		if(read_vector((int *)myvars->B,N)<0) {
			printf("File has negative arguments.\n");
			exit(0);
		}

		// compute how many clock ticks correspon to 1 sec
		ticspersec=(double)sysconf(_SC_CLK_TCK);

		// start timing for the parallel part
		t1=(double)times(&tb);

		// initialize mutex
		pthread_mutex_init(&dotsum, NULL);

		pthread_attr_init(&attr);
    
		// number of data assigned to each thread (assuming no remainder)
		stripSize= N/NPROC;

		myvars->DOTPRODUCT = 0;

		// create pthreads for computing dotproduct
		for(i=myid;i<NPROC;i++) {
				pthread_create(&dotThread[i],&attr,DotProduct,(void *)i);
		}


		//system ("sleep 5");

		// wait until all threads have terminated (barrier)
		for(i=myid;i<NPROC;i++) {
				pthread_join(dotThread[i],NULL);
		}
     
		printf("Final value - DOTPRODUCT:%d \n", myvars->DOTPRODUCT);
     
		// end time
		t2=(double)times(&tb);
     
     
		avrg_time+=(t2-t1)/ticspersec;
		sum_avrg_time_all_iterations += avrg_time;
          

	}
	// optional print
   /*
   for (i=0; i< N; i++) {
      if (i==0)
         printf(" A ");
      else
         printf("   ");
      for (j=0; j< N; j++)
        printf("%d ", myvars->A[i][j]);
      if (i==0)
         printf(" B ");
      else
         printf("   ");
      for (j=0; j< N; j++)
        printf("%d ", myvars->B[i][j]);
      if (i==0)
         printf(" C ");
      else
         printf("   ");
      for (j=0; j< N; j++)
        printf("%d ", myvars->C[i][j]);
      printf("\n");
   }
   */
	// results
	printf("Average time for (C=A*B): %lf sec.\n\n", sum_avrg_time_all_iterations/NO_ITERATIONS);

	free (myvars);
	pthread_mutex_destroy(&dotsum);

        return 0;
}
 
// function to compute DotProduct (called by every pthread)
void *DotProduct (void *arg)
{
	int *x, *y,i,first,last;
	long id = (long) arg;
	int localsum;
	  
	first = id*stripSize;
	last = first + stripSize-1;
      
	localsum = 0;
	x=myvars->A;
	y=myvars->B;
      
	for (i=first; i<=last ; i++) {
		localsum += (x[i] * y[i]);
	}

	pthread_mutex_lock (&dotsum);
		//printf("thread_id:%d adds localsum:%d to dotproduct:%d \n", id, localsum, myvars->DOTPRODUCT);
		myvars->DOTPRODUCT += localsum;
		//pthread_yield();
		//printf("thread_id:%d - DOTPRODUCT:%d \n", i, myvars->DOTPRODUCT);
	pthread_mutex_unlock (&dotsum);
      
} 
