// lock vs trylock (HAZARDS!)
//
// Read Section 11.6.1 from Rago & Stevens book, especially
// pthread_mutex_init kai  pthread_mutex_lock,  pthread_mutex_trlock,
// pthread_mutex_unlock ...
// 
// Program Description:							
// We assume two racers, at two diagonally opposite corner of a rectangular 
// region. They have to traverse along the roads along the peripheri of the 
// region. There are two bridges on two opposite sides of the rectangle. 
// In order to complete one round of traversal around this, the racers have   
// to get pass for both the bridge at a time. The conditions of the race are 
// 1) Only one racer can get a pass at a time. 
// 2) Before one starts one round, he has to request and get both the passes  
//    and then after finishing that round has to release the passes, and make  
//    new try to get those passes for the next round.			
// 3) Racer1 (R1) will acquire bridge-pass B1 first, then B2. 
//    R2 will acquire B2 and then B1. 
// 4) There is a maximum number of rounds prefixed. Whoever finishes that first, 
//    will be the winner and the race will stop. 
// 
// This is how the situation looks before starting. 
//	
// 
//		    	       B0 
//                 R0-------- ~  -------------
//		     |		     	      | 
//		     |			      | 
//		     |			      | 
//		     |			      | 
//		     --------- ~  ------------- R1
//			       B1 
// 
// The main function simulates two threads for two racers. 
// We avoid deadlock in calling natural B1/B2 and B2/B1 sequences of locks through trylock
// In this situation, if you acquire one lock and fail to acquire the other, then you must
// release the first lock and try again. 
//
// The problem with this plan is that it may not be possible to release the first lock safely.
// You might need the first lock to start an operation and the second lock to complete it.
// If you have to release the first lock, then you also have to undo the associated work.
// Effectively, you need to implement transactions.
//
// 
// gcc program.c -lpthread 
// ./a.out
// 
// LAB:	1) Examine different possibilities with commented code
//	2) Read about unlocking rules
//	3) Dining Philosophers in wikipedia (examine and play with solutions)

#define _GNU_SOURCE     // required by pthread_yield (do: man pthread_yield)

#include <stdio.h>  
#include <pthread.h> 
#include <stdlib.h> 
#include <stdint.h> 
#include <errno.h>

#define thread_num 2  
#define max_rounds 1000
#define TRUE  1 
#define FALSE 0
  
/* mutex locks for each bridge */  
pthread_mutex_t B0=PTHREAD_MUTEX_INITIALIZER, 
		B1=PTHREAD_MUTEX_INITIALIZER;
 
/* racer ID */ 
int myid=0;
  
/* number of rounds completed by each racer */  
int numRounds[thread_num]={0,0};  
  
void *racer(void *); /* prototype of racer routine */  
  
int main()  
{  
	pthread_t tid[thread_num];
	void *status=NULL; 
	int i,j;  
 
	/* create 2 threads representing 2 racers */  
	for (i = 0; i < thread_num; i++)  { 
		if (pthread_create(&tid[i], NULL, racer, (void *)(intptr_t) i) != 0)  
		{
			perror("pthread_create() failure \n");  
			exit(1);  
		}
	}
  
	for (i = 0; i < thread_num; i++)  
	{  
		if(!pthread_join(tid[i], &status)==0)  
		{  
			perror("pthread_join() failure \n");  
			exit(1);  
		}
	}   

	for (i=0; i<thread_num; i++)  
		printf("Racer %d finished %d rounds!!\n", i, numRounds[i]);  
     
	if (numRounds[0] >= numRounds[1]) 
		printf("\n 1ST RACER WINS..");  
	else
		printf("\n 2ND RACER WINS..");  
 
	return (0);  

}  
  
  
void *racer(void  *arg)  
{
	int index = (intptr_t) arg;
	int  i, NotYet; 
	float local_comp = 1;

	printf("Hello from Racer %d \n", index);
 
	// think: how to rewrite on exact number of rounds = max_rounds ?
	// usually it only approximates
	while ( numRounds[0] + numRounds [1] < max_rounds )  {
 
		NotYet = 1;

		/* RACER 0 tries to get both locks before she makes a round */  
		if (index==0) {
			while (NotYet) {

				// try to get the B0 lock first
				pthread_mutex_lock(&B0); 
				// retry using the following two possibilities
				// while (pthread_mutex_trylock(&B0)) ;  // is this the same as pthread_mutex_lock?
				// try single nonblocking call with pthread_mutex_trylock (see lines 134-136)
				printf("Racer %d got lock B0 & waiting for B1 \n", index); 
				fflush(stdout);

				// delay to compute (using usleep or pthread_yield => deadlock occurs more often)
				for (i=0; i<10; i++)
					local_comp = local_comp * 1.0753263713;

				/* try to get the B1 lock next */
				pthread_mutex_lock(&B1);
				NotYet = 0;

				/* less-greedy option: could not get B1 therefore unlock B0 */  
				//if (pthread_mutex_trylock(&B1) == EBUSY) { 
				//	pthread_mutex_unlock(&B0);
				//	// this gives more chances to other thread to acquire lock
				//	//usleep(1);
				//	//pthread_yield();
				//}
				//else
				//	NotYet = 0; /* got both locks  */
			}
       			printf("Racer %d got B0 and B1 \n",index);  
			fflush(stdout);
			// unlock both locks, not necessarily in reverse order
			pthread_mutex_unlock(&B0);  
			pthread_mutex_unlock(&B1);  
			if(lrand48()%2 == 1)  
				pthread_yield(); /* random yield to another thread */ 
		}
		/* RACER 1 tries to get both locks before she makes a round */  
		else if (index==1) {
			while (NotYet) {

				/* try to get the B1 lock first */
				pthread_mutex_lock(&B1);
				//while ( pthread_mutex_trylock(&B1) ) ;  // try to get the B0 lock first
				// try single nonblocking call with trylock (repeat lines 167-170)
            			printf("Racer %d got lock B1 & waiting for B0 \n", index); 
				//fflush(stdout);

				// delay to compute (using usleep or pthread_yield => deadlock occurs more often)
				for (i=0; i<10; i++)
					local_comp = local_comp * 1.0753263713;

            			/* try to get the B0 lock next */
				pthread_mutex_lock(&B0);
				NotYet = 0;

				/* less-greedy option: could not get B1 therefore unlock B0 */  
				//if (pthread_mutex_trylock(&B0) == EBUSY) { 
				//	pthread_mutex_unlock(&B1);
				//	// this gives more chances to other thread to acquire lock
				//	//usleep(1);
				//	//pthread_yield();
				//}
				//else
				//	NotYet = 0; /* got both locks  */
			}
       			printf("Racer %d got B0 and B1 \n",index);  
			fflush(stdout);
			// unlock both locks, not necessarily in reverse order
			pthread_mutex_unlock(&B1);  
			pthread_mutex_unlock(&B0);  
			if(lrand48()%2 == 1)  
				pthread_yield(); /* random yield to another thread */ 
		}
		numRounds[index]++;      /* Make one more round */
	}
	printf("Racer %d made %d rounds !\n", index, numRounds[index]);  
}
